// This program can be used to find a sequence of instructions which
// produce a desired constant.

import std.stdio;
import std.conv;

enum Operation: char {
    Increment = 'i',
    Decrement = 'd',
    Square = 'q',
    ShiftLeft = 'l',
    ShiftRight = 'r',
    Negate = '-',
}

// A handful of hand-optimized constants
immutable Operation[][] known = [
    cast(immutable Operation[]) /*  0 */ "",
    cast(immutable Operation[]) /*  1 */ "i",
    cast(immutable Operation[]) /*  2 */ "ii",
    cast(immutable Operation[]) /*  3 */ "iii",
    cast(immutable Operation[]) /*  4 */ "iil",
    cast(immutable Operation[]) /*  5 */ "iili",
    cast(immutable Operation[]) /*  6 */ "iiil",
    cast(immutable Operation[]) /*  7 */ "iiili",
    cast(immutable Operation[]) /*  8 */ "iill",
    cast(immutable Operation[]) /*  9 */ "iiiq",
    cast(immutable Operation[]) /* 10 */ "iilil",
    cast(immutable Operation[]) /* 11 */ "iilili",
    cast(immutable Operation[]) /* 12 */ "iiill",
    cast(immutable Operation[]) /* 13 */ "iiilli",
    cast(immutable Operation[]) /* 14 */ "iiilil",
    cast(immutable Operation[]) /* 15 */ "iilqd",
    cast(immutable Operation[]) /* 16 */ "iilq",
];

// https://en.wikipedia.org/wiki/Integer_square_root#Using_bitwise_operations
int sqrt(in int x) {
    if(x <= 0) return 0;
    int shift = 2;
    int shifted = x >> shift;
    while(shifted && shifted != x) {
        shift += 2;
        shifted = x >> shift;
    }
    shift -= 2;
    int result = 0;
    while(shift >= 0) {
        result = result << 1;
        int candidateResult = result + 1;
        if(candidateResult * candidateResult <= x >> shift) {
            result = candidateResult;
        }
        shift -= 2;
    }
    return result;
}

auto reverse(in int x, in Operation op) {
    struct Result {int value; bool allowed;}
    with(Operation) final switch(op) {
        case Increment: return Result(x - 1, x != int.min);
        case Decrement: return Result(x + 1, x != int.max);
        case Square:
            if(x < 0) return Result(0, false);
            const i = sqrt(x);
            return Result(i, i * i == x);
        case ShiftLeft: 
            const i = x >> 1;
            return Result(i, i << 1 == x);
        case ShiftRight: 
            const i = x << 1;
            return Result(i, i >> 1 == x);
        case Negate: return Result(-x, true);
    }
}

string opsToString(in Operation[] ops) {
    string result = "";
    foreach(op; ops) result ~= op;
    return result;
}

// Tree-like data structure used in finding a solution
struct FindNode{
    // Value of this node
    int val;
    // Depth of this node in the tree (root = 0)
    uint depth;
    // Parent node
    FindNode* prev;
    // The operation applied to the parent node to produce this one
    Operation op;
    
    // Get the list of operations which were executed on the goal
    // to arrive at the value represented by this node
    Operation[] ops() const {
        Operation[] list;
        const(FindNode)* node = &this;
        while(node !is null && node.prev !is null) {
            list ~= node.op;
            node = node.prev;
        }
        return list;
    }
}

// Search the operation tree for a matching solution
Operation[] find(in int goal, in int start = 0) {
    // Check for easy known solutions
    if(start == 0 && goal >= 0 && goal < known.length) {
        return known[goal].dup;
    }
    // Initialize the operations tree
    auto root = FindNode(goal, 0, null);
    FindNode[] queue = [root];
    enum TryOps = [
        Operation.Increment, Operation.Decrement,
        Operation.ShiftLeft, Operation.Square,
        Operation.ShiftRight, Operation.Negate,
    ];
    // Compute nodes for the operations tree breadth-first until
    // finding a node whose value is the expected starting value
    size_t i = 0;
    while(i < queue.length) {
        auto node = queue[i];
        foreach(tryOp; TryOps) {
            auto rev = reverse(node.val, tryOp);
            if(rev.allowed && start == 0 && rev.value >= 0 && rev.value < known.length) {
                return known[rev.value] ~ tryOp ~ node.ops;
            }else if(rev.allowed && rev.value == start) {
                return tryOp ~ node.ops;
            }else if(rev.allowed && node.depth <= 8) {
                auto next = FindNode(rev.value, 1 + node.depth, &queue[i], tryOp);
                queue ~= next;
            }
        }
        i++;
    }
    // Failed to find any solution
    return [];
}

int main(in string[] args) {
    if(args.length < 2) {
        writeln("Usage: find_constant <goal> <start? = 0>");
        writeln("Example: find_constant 1234 # Set A to 1234 (starting at 0)");
        writeln("Example: find_constant 999 123 # Set A to 999 (starting at 123)");
        return 0;
    }
    writeln("Working...");
    const int goal = args[1].to!int;
    const int start = args.length > 2 ? args[2].to!int : 0;
    auto result = find(goal, start);
    if(start == 0) {
        if(result.length) {
            writeln("Found code to produce constant ", goal, ":");
            writeln(result.opsToString);
            return 1;
        }else {
            writeln("Failed to find code to produce constant ", goal, ".");
            return 0;
        }
    }else {
        auto fromZero = find(goal, 0);
        writeln("Found code to produce constant ", goal, " starting from value ", start, ":");
        if(fromZero.length && (result.length == 0 || fromZero.length < result.length)) {
            writeln('0', fromZero.opsToString);
            return 0;
        }else if(result.length) {
            writeln(result.opsToString);
            return 0;
        }else {
            writeln("Failed to find code to produce constant ", goal, " starting from value ", start, ".");
            return 1;
        }
    }
}
