import std.array;
import std.conv;
import std.datetime;
import std.file;
import std.process;
import std.stdio;

// Represent test cases loaded from a single file in the tests/ directory
struct Test {
    string path;
    string[] inputs;
    string[] outputs;
    string source;
    bool passed;
}

// Look for the Shark interpreter binary
string getBinary() {
    enum tryPaths = [
        "bin/shark.exe",
        "bin/shark",
        "shark.exe",
        "shark",
    ];
    foreach(path; tryPaths) {
        if(path.exists) return path;
    }
    return "";
}

// Do the things
int main(in string[] args) {
    // Find binary (or accept path as first argument)
    string testsPath = args.length > 1 ? args[1] : "tests";
    string binaryPath = args.length > 2 ? args[2] : getBinary();
    if(!binaryPath || !binaryPath.exists) {
        writeln("Unable to find Shark binary.");
        writeln("Try passing the binary path as an argument.");
        writeln("Example: run_tests path/to/tests path/to/shark");
        return 1;
    }
    // Find all *.ck files in the tests/ directory
    Test[] tests;
    const timestamp = Clock.currTime().toISOString();
    auto entries = dirEntries(testsPath, "*.sk", SpanMode.shallow);
    foreach(filePath; entries) {
        Test test = Test(filePath);
        uint state = 0;
        foreach(line; File(filePath).byLine) {
            if(line == "##INPUT") { // Provided input
                test.inputs ~= ""; state = 1;
            }else if(line == "##OUTPUT") { // Expected output
                test.outputs ~= ""; state = 2;
            }else if(line == "##SOURCE") { // Shark source
                state = 3;
            }else if(state == 1) {
                test.inputs[$ - 1] ~= line ~ "\n";
            }else if(state == 2) {
                test.outputs[$ - 1] ~= line ~ "\n";
            }else if(state == 3) {
                test.source ~= line ~ "\n";
            }
        }
        if(test.outputs.length && test.source.length) {
            tests ~= test;
        }
    }
    // Now run the tests
    writeln("Running ", tests.length, " tests:");
    uint passedCount = 0;
    // For each test...
    foreach(test; tests) {
        test.passed = true;
        // For each expected output (e.g. given different inputs)...
        for(size_t i = 0; i < test.outputs.length; i++) {
            // Write temporary input and output files
            const istr = to!string(1 + i);
            string inputPath = test.path ~ "." ~ istr ~ "." ~ timestamp ~ ".in.log";
            string outputPath = test.path ~ "." ~ istr ~ "." ~ timestamp ~ ".out.log";
            auto inputFile = File(inputPath, "wb");
            auto outputFile = File(outputPath, "wb");
            inputFile.write(i < test.inputs.length ? test.inputs[i] : "");
            inputFile.open(inputPath, "rb");
            // Run Shark interpreter with given input, output, and source
            char[4096] outputBuffer;
            auto process = spawnProcess(
                [binaryPath, "eval", test.source], inputFile, outputFile
            );
            wait(process);
            // Cleanup
            inputFile.close();
            inputPath.remove();
            outputFile.open(outputPath, "rb");
            char[] output = outputFile.rawRead(outputBuffer);
            outputFile.close();
            // Remove output log on successful test
            if(output == test.outputs[i] || output.replace("\r\n", "\n") == test.outputs[i]) {
                writeln("Test passed: ", test.path, " (input #", istr, ")");
                outputPath.remove();
            // Report failures and keep the output log, for debugging purposes
            }else {
                writeln("TEST FAILURE: ", test.path, " (input #", i + 1, ")");
                test.passed = false;
            }
        }
        passedCount += test.passed;
    }
    // All done!
    writeln(passedCount, " of ", tests.length, " tests passed.");
    return passedCount == tests.length ? 0 : 1;
}
