import std.array;
import std.bigint;
import std.file;
import std.stdio;
import std.string;
import std.utf;

import core.stdc.stdio : FILE, fgetc;

int main(in string[] args) {
    // shark eval <source>
    if(args.length > 2 && args[1] == "eval") {
        string src = args[2];
        for(size_t i = 3; i < args.length; i++) src ~= args[i];
        return Program.loadSource(src).execute();
    // shark load <file path>
    }else if(args.length > 2 && args[1] == "load") {
        const status = Program.loadFile(args[2]).execute();
        if(status) writeln("Failed to load file from path \"", args[2], "\".");
        return status;
    // shark <file path>
    }else if(args.length > 1) {
        const status = Program.loadFile(args[1]).execute();
        if(status) writeln("Failed to load file from path \"", args[1], "\".");
        return status;
    // Otherwise show usage
    }else {
        writeln("Usage: shark <file> | shark load <file> | shark eval <code>");
        writeln("Example: shark example.sk");
        return 0;
    }
}

// Enumeration of recognized instructions
enum Instruction: char {
    // Do nothing
    Nop = 'z',
    // Debug helper; dump info to console
    DebugDump = 'D',
    // Jump over the following instruction when A is zero
    SkipIfZero = '?',
    // Jump over the following instruction when A is nonzero
    SkipIfNonzero = '!',
    // Jump back 3 instructions
    JumpBack = '{',
    // Add the current location of the instruction counter to the control stack
    PushCallStack = '^',
    // Pop control stack and resume execution after the ^
    PopCallStack = '~',
    // Resume execution after the ^ on the top of the control stack, but don't pop
    PeekCallStack = '&',
    // Pop the control stack but do not jump to the corresponding ^
    DiscardCallStack = 'x',
    // Set memory at address A to the value of B
    StoreMemory = '>',
    // Set B to the value in memory at address A
    LoadMemory = '<',
    // Swap the values of B and the value in memory at address A
    SwapMemory = 'w',
    // Output A to the console as a decimal number
    WriteNumber = ':',
    // Output A to the console as a unicode code point (Write nothing if not a valid code point or surrogate)
    WriteUnicode = ';',
    // Output a newline to the console
    WriteNewline = 'n',
    // Wait for a line of console input, interpret as a number, and store in A
    // Set A to 0 and B to 0 if a number couldn't be read
    ReadNumber = '.',
    // Wait for a unicode code point in console input and store in A
    ReadUnicode = ',',
    // Swap the values of A and B
    SwapAB = '@',
    // Swap the values of A and C
    SwapAC = '\'',
    // Swap the values of B and D
    SwapBD = '"',
    // Copy the value of A into B
    CopyAB = '$',
    // Set A to zero
    ZeroRegister = '0',
    // Increment A
    Increment = 'i',
    // Decrement A
    Decrement = 'd',
    // Square A
    Square = 'q',
    // Shift bits of A left once
    ShiftLeft = 'l',
    // Shift bits of A right once (Reduce magnitude and retain sign)
    ShiftRight = 'r',
    // Negate A
    Negate = '-',
    // Assign A to the sum of A and B
    Sum = '+',
    // Assign A to the product of A and B
    Product = '*',
    // Assign A to the remainder of A divided by B (Terminate when B is zero)
    Remainder = '%',
}

// Compensate for deficient standard library...
struct FileRange {
    FILE* file;
    int value = int.min;
    
    this(File file) {
        this.file = file.getFP();
    }
    
    void next() {
        this.value = fgetc(this.file);
    }
    
    bool empty() {
        if(this.value == int.min) this.next();
        return this.value < 0;
    }
    
    char front() {
        if(this.value == int.min) this.next();
        assert(this.value >= 0);
        return cast(char) this.value;
    }
    
    void popFront() {
        this.value = int.min;
    }
}

string intToString(BigInt value) {
    string result = "";
    void sink(in char[] chars) {result ~= chars;}
    value.toString(&sink, "%d");
    return result;
}

// Interpreter data structure
struct Program {
    // Flag set when there was a fatal error
    // Currently the only fatal error is failure to load a file...
    bool error;
    // List of instructions to execute, parsed from source code
    Instruction[] src;
    // Value of register A
    BigInt A;
    // Value of register B
    BigInt B;
    // Value of register C
    BigInt C;
    // Value of register D
    BigInt D;
    // Most recently accessed memory address (used by debug output)
    BigInt lastMemoryAddress;
    // Program counter (execution location in source)
    size_t pc = 0;
    // Call stack (as an array)
    size_t[] controlStack;
    // Program memory (as a map/associative array)
    BigInt[BigInt] memory;
    
    // Determine if a char in source is a recognized instruction
    static bool isInstruction(in char ch) {
        foreach(instruction; __traits(allMembers, Instruction)) {
            mixin(`if(ch == Instruction. ` ~ instruction ~ `) return true;`);
        }
        return false;
    }
    
    // Load source from file path
    static Program loadFile(in string path) {
        string src = void;
        try {
            src = cast(string) read(path);
        }catch(Exception exception) {
            return Program(true);
        }
        return Program.loadSource(src);
    }
    
    // Load source code
    static Program loadSource(in string src) {
        Instruction[] instructions;
        instructions.reserve(src.length);
        bool inComment = false;
        foreach(ch; src) {
            if(ch == '#') {
                inComment = true;
            }else if(ch == '\n') {
                inComment = false;
            }else if(!inComment && isInstruction(ch)) {
                instructions ~= cast(Instruction) ch;
            }
        }
        return Program(false, instructions);
    }
    
    // Execute the program
    int execute() {
        if(this.error) {
            return 1;
        }else if(this.src.length) {
            while(!this.next()) {}
        }
        return 0;
    }
    
    // Pop the control stack and return the popped value
    size_t popCallStack() {
        assert(this.controlStack.length);
        const size_t value = this.controlStack[$ - 1];
        this.controlStack.length--;
        return value;
    }
    
    // Get the value at a memory address (get 0 if not yet assigned)
    BigInt loadMemory(in BigInt address) {
        this.lastMemoryAddress = address;
        if(BigInt* value = address in this.memory) return *value;
        else return BigInt(0);
    }
    
    // Evaluate the next instruction
    // Returns true when the program was terminated
    bool next() {
        if(this.pc >= this.src.length) {
            return true;
        }
        with(Instruction) switch(this.src[this.pc]) {
            case Nop:
                this.pc++;
                return false;
            case DebugDump:
                writeln("DEBUG DUMP @ PC ", this.pc);
                writeln("A: ", intToString(this.A));
                writeln("B: ", intToString(this.B));
                writeln("C: ", intToString(this.C));
                writeln("D: ", intToString(this.D));
                writeln("mem[", this.lastMemoryAddress, "]: ",
                    this.loadMemory(this.lastMemoryAddress)
                );
                this.pc++;
                return false;
            case SkipIfZero:
                this.pc += this.A == 0 ? 2 : 1;
                return false;
            case SkipIfNonzero:
                this.pc += this.A == 0 ? 1 : 2;
                return false;
            case JumpBack:
                if(this.pc < 3) return true;
                this.pc -= 3;
                return false;
            case PushCallStack:
                this.controlStack ~= this.pc++;
                return false;
            case PopCallStack:
                if(!this.controlStack.length) return true;
                this.pc = 1 + this.popCallStack();
                return false;
            case PeekCallStack:
                if(!this.controlStack.length) return true;
                this.pc = 1 + this.controlStack[$ - 1];
                return false;
            case DiscardCallStack:
                if(!this.controlStack.length) return true;
                this.popCallStack();
                this.pc++;
                return false;
            case StoreMemory:
                this.memory[this.A] = this.B;
                this.lastMemoryAddress = this.A;
                this.pc++;
                return false;
            case LoadMemory:
                this.B = this.loadMemory(this.A);
                this.pc++;
                return false;
            case SwapMemory:
                const temp = this.B;
                this.B = this.loadMemory(this.A);
                this.memory[this.A] = temp;
                this.pc++;
                return false;
            case WriteNumber:
                write(intToString(this.A));
                this.pc++;
                return false;
            case WriteUnicode:
                const codePoint = this.A.toLong();
                const isValid = codePoint >= 0 && codePoint < 0x110000;
                if(isValid) write(cast(dchar) codePoint);
                this.pc++;
                return false;
            case WriteNewline:
                write('\n');
                this.pc++;
                return false;
            case ReadNumber:
                try {
                    this.A = BigInt(readln().strip());
                }catch(Exception exception) {
                    this.B = 0;
                }
                this.pc++;
                return false;
            case ReadUnicode:
                auto chars = FileRange(stdin).byUTF!dchar();
                const dchar front = chars.front;
                this.A = BigInt(cast(uint) front);
                this.pc++;
                return false;
            case SwapAB:
                const temp = this.A;
                this.A = this.B;
                this.B = temp;
                this.pc++;
                return false;
            case SwapAC:
                const temp = this.A;
                this.A = this.C;
                this.C = temp;
                this.pc++;
                return false;
            case SwapBD:
                const temp = this.B;
                this.B = this.D;
                this.D = temp;
                this.pc++;
                return false;
            case CopyAB:
                this.B = this.A;
                this.pc++;
                return false;
            case ZeroRegister:
                this.A = 0;
                this.pc++;
                return false;
            case Increment:
                this.A++;
                this.pc++;
                return false;
            case Decrement:
                this.A--;
                this.pc++;
                return false;
            case Square:
                this.A *= this.A;
                this.pc++;
                return false;
            case ShiftLeft:
                this.A <<= 1;
                this.pc++;
                return false;
            case ShiftRight:
                this.A >>= 1;
                this.pc++;
                return false;
            case Negate:
                this.A = -this.A;
                this.pc++;
                return false;
            case Sum:
                this.A += this.B;
                this.pc++;
                return false;
            case Product:
                this.A *= this.B;
                this.pc++;
                return false;
            case Remainder:
                if(this.B == 0) return true;
                this.A %= this.B;
                this.pc++;
                return false;
            default:
                assert(false);
        }
    }
}

